/*	
 *	usbusr.c
 *
 *	USB [U[vO iUSB Xg[Wj
 *	Copyright (c) 1997 Philips Semiconductors
 *	Copyright (c) 2002 Naoyuki Sawa
 *
 *	* Mon Sep 12 22:31:00 JST 2002 Naoyuki Sawa
 *	- 쐬JnB
 */
#include <piece.h>
#include <s1c33cpu.h>
#include "c33209.h"
#include "d12ci.h"
#include "usbsys.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void standard_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
void class_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
void ep2_rxdone(int st);
void ep2_txdone(int st);
USBINFO ui = {
	/* USBC mode */
	D12_ENDP_NONISO,	/* mode */
	/* USBC raw event */
	NULL,			/* bus_reset */
	NULL,			/* ep1_rxdone */
	NULL,			/* ep1_txdone */
	ep2_rxdone,		/* ep2_rxdone */
	ep2_txdone,		/* ep2_txdone */
	/* Device request */
	standard_request,	/* standard_request */
	class_request,		/* class_request */
	NULL,			/* vendor_request */
};

/****************************************************************************
 *	fBXNv^
 ****************************************************************************/

static unsigned char DeviceDesc[18] = {
	0x12,			// bLength		
	0x01,			// bDescriptorType	foCXfBXNv^
	0x10,0x01,		// bcdUSB		USB Version 1.1
	0x00,			// bDeviceClass		eC^[tF[Xɐݒ
	0x00,			// bDeviceSubClass	gp
	0x00,			// bDeviceProtocol	gp
	EP0_PACKET_SIZE,	// bMaxPacketSize0	EP0̍őpPbgTCY
	0x19,0x0e,		// idVendor		PCE_VENDOR_ID
	0x13,0x10,		// idProduct		PCE_PRODUCT_ID+0x13
	0x00,0x01,		// bcdDevice		foCX[Xԍ 1.00
	0x00,			// iManufacture		gp
	0x00,			// iProduct		gp
	0x00,			// iSerialNumber	gp
	0x01			// bNumConfiguraions	RtBO[V1
};
static unsigned char ConfigDesc[9] = {
	0x09,			// bLength		
	0x02,			// bDescriptorType	RtBO[VfBXNv^
	0x20,0x00,		// wTotalLength		CFG(9)+I/F(9)+EP(7*2)=32=0x20
	0x01,			// bNumberIntefaces	C^[tF[X1
	0x01,			// bConfigurationValue	RtBO[V1Ȃ̂#1̂
	0x00,			// iConfiguration	gp
	0x80,			// bmAttributes		oXp[foCXA[gEFCNAbv
	0x32			// MaxPower		d 100mA (100/2=50)
};
static unsigned char InterfaceDesc[9] = {
	0x09,			// bLength		
	0x04,			// bDescriptorType	C^[tF[XEfBXNv^
	0x00,			// bInterfaceNumber	C^[tF[X#0
	0x00,			// bAlternateSetting	I^l[gZbg͖
	0x02,			// bNumEndpointes	Gh|Cg2
	0x08,			// bInterfaceClass	Mass storage class
	0x06,			// bInterfaceSubClass	SCSI transparent
	0x50,			// bInterfaceProtocol	Bulk only
	0x00			// iInterface		gp
};
static unsigned char EndpointDesc[2][7] = {
{
	0x07,			// bLength		
	0x05,			// bDescriptorType	Gh|CgfBXNv^
	0x82,			// bEndpointAddress	oNIN
	0x02,			// bAttributes		oN]
	EP2_NONISO_PACKET_SIZE,0x00, // wMaxPacketize	]oCg
	0x00,			// bInterval		C^[o[ms] (gp)
},{
	0x07,			// bLength		
	0x05,			// bDescriptorType	Gh|CgfBXNv^
	0x02,			// bEndpointAddress	oNOUT
	0x02,			// bAttributes		oN]
	EP2_NONISO_PACKET_SIZE,0x00, // wMaxPacketize	]oCg
	0x00,			// bInterval		C^[o[ms] (gp)
}
};

static unsigned char*
dscpy(unsigned char *ptr, const void *src, int len)
{
	memcpy(ptr, src, len);
	return ptr + len;
}

static unsigned char tmpbuff[256]; /* SfBXNv^[ȃTCYKv */

/****************************************************************************
 *	USB Xg[W
 ****************************************************************************/

#define HI(w)	((w) >> 8 & 0xff)
#define LO(w)	((w) >> 0 & 0xff)

#ifndef BLOCKSIZE
#define BLOCKSIZE		4096		/* tbṼZN^TCY */
#endif /*BLOCKSIZE*/
#define DEVNUL	((unsigned char*)0x1000000)	/* S~f[^MpAhX */

#define SCTSIZ	BLOCKSIZE			/* ZN^TCY            */
#define SCTCNT	48				/* SZN^              */
#define CLSCNT	(SCTCNT - 3)			/* NX^              */
#define CLSMIN	2				/* NX^ԍŏl      */
#define CLSMAX	(CLSMIN + CLSCNT - 1)		/* NX^ԍől      */
#define DIRCNT	(SCTSIZ / 32)			/* fBNgGg  */
unsigned char scts[SCTCNT][SCTSIZ];		/* fBXNC[W        */
int dirty;					/* IɕۑKvȂ1 */
unsigned rx_cnt;				/* MoCg            */
unsigned tx_cnt;				/* MoCg            */
unsigned err_cnt;				/* G[          */

#define USBSTOR_FILENAME	"usbstor.img"	/* fBXNC[Wۑpt@C */
#define USBSTOR_FILESIZE	(SCTSIZ*(SCTCNT-1)) /* ۑpt@CTCY(u[gZN^͕ۑȂ) */

FAT_BOOT_SECTOR fat_boot_sector = {
	{0xeb,0x3c,0x90},			/* ignored[3] */
	{'P','I','E','C','E','-','M','E'},	/* system_id[8] */
	{LO(SCTSIZ),HI(SCTSIZ)},		/* sector_size[2] */
	1,					/* cluster_size */
	1,					/* reserved */
	1,					/* fats */
	{LO(DIRCNT),HI(DIRCNT)},		/* dir_entries[2] */
	{LO(SCTCNT),HI(SCTCNT)},		/* sectors[2] */
	0xf8,					/* media */
	1,					/* fat_length */
	1,					/* secs_track */
	1,					/* heads */
	0,					/* hidden */
};

#define PHASE_IDLE	0		/* CBWM҂      */
#define PHASE_RX	1		/* ǉf[^M */
#define PHASE_TX	2		/* ǉf[^M */
#define PHASE_CSW	3		/* CSWM҂      */

int phase;				/* tF[Y(PHASE_*)          */
CBW cbw;				/* CBWMp                  */
CSW csw;				/* CSWM\p              */
unsigned char* data_ptr;		/* Mf[^̐擪AhX */
int data_cnt;				/* Mς݃f[^̃oCg */

unsigned char inquiry_data[36] = {	/* INQUIRY f[^ */
	0x00,	/* 0: _CNgANZXfoCX */
	0x80,	/* 1: [oufBXN */
	0x02,	/* 2: o[WH0x02Œ */
	0x02,	/* 3: f[^`H0x02Œ */
	0x1f,	/* 4: ȍ~̃oCg   */
      /*  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,*/
	0x00,0x00,0x00,							/*  5- 7: 낢   */
	'P','I','E','C','E','L','A','B',				/*  8-15: x_       */
	'U','S','B','R','E','M','O','V','A','B','L','E','D','I','S','K',/* 16-31: i         */
	'0','.','0','1',						/* 32-35: o[Wԍ */
};

unsigned char capacity_data[8] = {	/* READ CAPACITY f[^ */
	0x00,0x00,HI(SCTCNT-1),LO(SCTCNT-1),	/* 0-3: ŏIZN^CfbNX(ZN^ł͂Ȃ!) */
	0x00,0x00,HI(SCTSIZ  ),LO(SCTSIZ  ),	/* 4-7: 1ZN^̃oCg                     */
};

unsigned char sense_data[18] = {	/* REQUEST SENSE f[^ */
	0x70,				/*     0: RESPONSE CODE = Current errors  */
	0x00,				/*     1: OBSOLATE                        */
	0x05,				/*     2: SENSE KEY = Illegal request     */
	0x00,0x00,0x00,0x00,		/*  3- 6: INFORMATION                     */
	0x0a,				/*     7: ADDITIONAL SENSE LENGTH         */
	0x00,0x00,0x00,0x00,		/*  8-11: COMMAND-SPECIFIC INFORMATION    */
	0x00,				/*    12: ADDITIONAL SENSE CODE           */
	0x00,				/*    13: ADDITIONAL SENSE CODE QUALIFIER */
	0x00,				/*    14: FIELD REPLACEMENT UNIT CODE     */
	0x00,0x00,0x00,			/* 15-17: SENSE-KEY SPECIFIC              */
};

/****************************************************************************
 *	WNGXg
 ****************************************************************************/

/* * AUDIO  HID NXł́ÃNGXg͗Ȃ̂ŁAȂĂ܂B
 * * MASS STORAGE NXł ENDPOINT_HALT NÂ߂ɎĝŎĂ݂܂B
 *   ܂mFĂ܂BmFOKB肭Ă悤łB
 */
static void
clear_feature(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	int endp;

	switch(req->bmRequestType & 0x1f) { /* recipient */
	case 0x00: /* device */
		switch(req->wValue) {
		case 1: /* device remote wakeup */
			break; /* T|[g */
		case 2: /* test mode */
			break; /* T|[g */
		}
		break;
	case 0x01: /* interface */
		/* no features */
		break;
	case 0x02: /* endpoint */
		switch(req->wValue) {
		case 0: /* endpoint halt */
			endp = req->wIndex & 0x7f;
			if(endp > 2) break; /* EPx */
			endp *= 2; /* EP0=0or1/EP1=2or3/EP2=4or5 */
			if(req->wIndex & 0x80) endp += 1; /* EPx IN */
			D12_SetEndpointStatus(endp, 0); /* STALL */
			usb_ep0_tx(NULL, 0);
			/*{{USB MASS STORAGE BULK-ONLY*/
			/* BULK-IÑXg[ƂɁAM҂CSWAɑM܂B */
			if(endp == 5 && phase == PHASE_CSW) ep2_txdone(0);
			/*}}USB MASS STORAGE BULK-ONLY*/
			return;
		}
		break;
	}

	usb_ep0_stall();
}

static void
set_address(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	D12_SetAddressEnable(req->wValue, 1);
	usb_ep0_tx(NULL, 0);
}

static void
get_descriptor(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	unsigned char* p;

	switch(req->wValue & 0xff00) {
	case 0x0100: /* Device */
		/* TODO: */
		usb_ep0_tx(DeviceDesc, sizeof DeviceDesc);
		return;
	case 0x0200: /* Configuration */
		/* TODO: */
		p = tmpbuff;
		p = dscpy(p, ConfigDesc, sizeof ConfigDesc);
		p = dscpy(p, InterfaceDesc, sizeof InterfaceDesc);
		p = dscpy(p, EndpointDesc, sizeof EndpointDesc);
		usb_ep0_tx(tmpbuff, p - tmpbuff);
		return;
	case 0x0300: /* String */
	case 0x0400: /* Interface */
	case 0x0500: /* Endpoint */
	case 0x0600: /* Device qualifier */
	case 0x0700: /* Other speed configuration */
	case 0x0800: /* Interface power */
		break; /* Ή */
	}

	usb_ep0_stall();
}

static void
set_configuration(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	D12_SetEndpointEnable(req->wValue != 0);
	usb_ep0_tx(NULL, 0);
}

static void
set_interface(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	usb_ep0_tx(NULL, 0);
}

static void (*standard_request_table[])(const DEVICE_REQUEST* req, const unsigned char* dat, int len) = {
/* 0*/	NULL/*get_status*/,
/* 1*/	clear_feature,
/* 2*/	NULL,
/* 3*/	NULL/*set_feature*/,
/* 4*/	NULL,
/* 5*/	set_address,
/* 6*/	get_descriptor,
/* 7*/	NULL/*set_descriptor*/,
/* 8*/	NULL/*get_configuration*/,
/* 9*/	set_configuration,
/*10*/	NULL/*get_interface*/,
/*11*/	set_interface,
/*12*/	NULL/*synch_frame*/,
};
#define N_STANDARD_REQUEST_TABLE (sizeof standard_request_table / sizeof *standard_request_table)

/****************************************************************************
 *
 ****************************************************************************/

void
standard_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	if(req->bRequest >= N_STANDARD_REQUEST_TABLE ||
	   standard_request_table[req->bRequest] == NULL) {
		usb_ep0_stall();
		return;
	}
	standard_request_table[req->bRequest](req, dat, len);
}

/* MASS STORAGE BULK-ONLYL̃NXNGXgB0xffn܂邱ƂɒӁI */
void
class_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	switch(req->bRequest) {
	case 0xff: /* Bulk-Only Mass Storage Reset */
		/* tF[YBXg[Ԃ͈ێ܂B */
		phase = PHASE_IDLE;
		usb_ep0_tx(NULL, 0);
		return;
	case 0xfe: /* Get Max LUN */
		/* PLUN̏ꍇSTALLOKB(Bulk-Onlydl) */
		break;
	}
	usb_ep0_stall();
}

void
ep2_rxdone(int st)
{
	int op, pos, len;

	switch(phase) {
	case PHASE_IDLE: /* CBWM҂ */
		/* CBWM܂B */
		len = D12_ReadEndpoint(4, &cbw, CBW_SIZE);
		if(len != CBW_SIZE) goto PHASE_ERROR; /* CBWs */
		if(cbw.dCBWSignature != CBW_SIGNATURE) goto PHASE_ERROR; /* CBWs */
		/* CSẄꕔĂ܂B */
		csw.dCSWSignature = CSW_SIGNATURE;
		csw.dCSWTag = cbw.dCBWTag;
		data_cnt = 0; /* JE^NAĂ܂ */
		/* R}hɂ... */
		op = cbw.CBWCB[0];
		switch(op) {
		case 0x00: /* TEST UNIT READY */
			len = 0;
			goto GOOD;
		case 0x03: /* REQUEST SENSE */
			/* zg͓K؂ȃZXL[ݒ肷ׂłA蔲ďIllegalRequestԂĂ܂B
			 * ȂƂLinuxJ[lł́AREQUEST SENSE̓eăG[ςĂ͂Ȃ悤łB
			 */
			data_ptr = sense_data;
			len = sizeof sense_data;
			goto GOOD;
		case 0x12: /* INQUIRY */
			/* zg͂ǂ̂悤ȏ񂪗vĂ邩𒲂ׂȂ΂Ȃ̂łA
			 *  WindowsLinuxł͕Wf[^ȊOvȂ悤Ȃ̂ŁA蔲Ă܂B
			 */
			data_ptr = inquiry_data;
			len = sizeof inquiry_data;
			goto GOOD;
		case 0x1b: /* START STOP UNIT */
			/* Windowsł͗Ȃ݂łB */
			len = 0;
			goto GOOD;
		case 0x1e: /* PREVENT ALLOWED REMOVAL */
			/* Windowsł͗Ȃ݂łB */
			len = 0;
			goto GOOD;
		case 0x25: /* READ CAPACITY */
			data_ptr = capacity_data;
			len = sizeof capacity_data;
			goto GOOD;
		case 0x28: /* READ(10) */
		case 0x2a: /* WRITE(10) */
			pos = cbw.CBWCB[2] << 24 |
			      cbw.CBWCB[3] << 16 |
			      cbw.CBWCB[4] <<  8 |
			      cbw.CBWCB[5] <<  0;
			len = cbw.CBWCB[7] <<  8 |
			      cbw.CBWCB[8] <<  0;
			if(pos < 0 || SCTCNT < pos + len) goto COMMAND_FAILED; /* ͈͊O */
			if(op == 0x2a/*WRITE*/ && pos == 0) goto COMMAND_FAILED; /* u[gZN^ی */
			if(op == 0x2a/*WRITE*/) dirty = 1; /* vۑ */
			data_ptr = &scts[pos][0];
			len *= SCTSIZ;
			goto GOOD;
		}
		/* dvI
		 * * T|[gR}hւ̉PHASE ERRORł͂ȂCOMMAND FAILEDłI
		 *   PHASE ERRORŉĂ܂ƃfoCXƃzXg̃tF[YȂȂA
		 *   zXg̃vOnOAbvACPUgp100%ɂȂ肵܂B
		 * * ȂȂȂ邩ƂƁAPHASE ERROROUT/INpCvXg[܂A
		 *   zXgINpCṽXg[REQUEST SENSE𔭍sĂ܂悤Ȃ̂łB
		 *   foCXREQUEST SENSEMłAzXg͂̂܂܉҂ɓĂ܂܂B
		 *   PHASE ERROROUT/INpCvXg[邱ƂԈႢH
		 * * ̂悤ȏǏ\h邽߂ɂ́uɗ́AXg[͎gȂvǂłB
		 *   R}hG[̏ꍇAiRłjf[^̑M̓zXg̗vʂɍsA
		 *   f[^MCSW.dCBWDataTransferLengthŃf[^RłƂʒmA
		 *   csw.bCSWStatus=COMMAND FAILEDŃG[ƒm点܂B
		 *   ƍlč΃Xg[Ă삷͂łA
		 *   lRf[^𑗎MĂ܂ȒPAŖȂ悤łB
		 * * ̃vOłAPHASE ERRORŉꍇ̋͗\ł܂B
		 *   T|[gR}ĥ悤ȕʂɗL蓾G[ɂCOMMAND FAILEDŉA
		 *   ǂ悤ȂȂꍇ̂PHASE ERRORŉ邱Ƃɂ܂B
		 */
		goto COMMAND_FAILED;
GOOD:
		if(cbw.bmCBWFlags & 0x80) { /* DATA-IN (device to host) */
			/* Hi>=DiK{BHi>Di̎HĩTCY܂ŃpfBOB */
			if(!(cbw.dCBWDataTransferLength >= len)) goto COMMAND_FAILED;
		} else { /* DATA-OUT (host to device) */
			/* Ho=DoK{B */
			if(!(cbw.dCBWDataTransferLength == len)) goto COMMAND_FAILED;
		}
		csw.dCSWDataResidue = cbw.dCBWDataTransferLength - len;
		csw.bCSWStatus = 0x00; /* GOOD */
		if(cbw.bmCBWFlags & 0x80) { /* DATA-IN (device to host) */
			if(cbw.dCBWDataTransferLength) { /* Mf[^ */
				phase = PHASE_TX;
				ep2_txdone(0);
			} else { /* Mf[^Ȃ */
				phase = PHASE_CSW;
				ep2_txdone(0);
			}
		} else { /* DATA-OUT (host to device) */
			if(cbw.dCBWDataTransferLength) { /* Mf[^ */
				phase = PHASE_RX;
			} else { /* Mf[^Ȃ */
				phase = PHASE_CSW;
				ep2_txdone(0);
			}
		}
		return;
COMMAND_FAILED:
		data_ptr = DEVNUL; /* Mf[^͑SăS~ */
		csw.dCSWDataResidue = cbw.dCBWDataTransferLength;
		csw.bCSWStatus = 0x01; /* COMMAND FAILED */
		if(cbw.bmCBWFlags & 0x80) { /* DATA-IN (device to host) */
			if(cbw.dCBWDataTransferLength) { /* Mf[^ */
				phase = PHASE_TX;
				ep2_txdone(0);
			} else { /* Mf[^Ȃ */
				phase = PHASE_CSW;
				ep2_txdone(0);
			}
		} else { /* DATA-OUT (host to device) */
			if(cbw.dCBWDataTransferLength) { /* Mf[^ */
				phase = PHASE_RX;
			} else { /* Mf[^Ȃ */
				phase = PHASE_CSW;
				ep2_txdone(0);
			}
		}
		/*{{MO*/
		err_cnt++;
		/*}}MO*/
		return;
PHASE_ERROR:
		csw.dCSWDataResidue = cbw.dCBWDataTransferLength;
		csw.bCSWStatus = 0x02; /* PHASE ERROR */
		D12_SetEndpointStatus(4, 1);
		D12_SetEndpointStatus(5, 1);
		/* BULK-IÑXg[ɁACSW𑗐M悤\񂵂܂B */
		phase = PHASE_CSW;
		/*{{MO*/
		err_cnt++;
		/*}}MO*/
		return;

	case PHASE_RX: /* f[^M */
		len = cbw.dCBWDataTransferLength - data_cnt;
		if(len > EP2_NONISO_PACKET_SIZE) len = EP2_NONISO_PACKET_SIZE;
		len = D12_ReadEndpoint(4, data_ptr + data_cnt, len);
		data_cnt += len;
		if(data_cnt == cbw.dCBWDataTransferLength) { /* M */
			/* Mobt@͋󂢂Ă͂Ȃ̂ŁACSW𑗐M܂B */
			phase = PHASE_CSW;
			ep2_txdone(0);
		}
		/*{{MO*/
		rx_cnt += len;
		/*}}MO*/
		return;
	}
}

void
ep2_txdone(int st)
{
	int len;

	switch(phase) {
	case PHASE_TX: /* f[^M */
		len = cbw.dCBWDataTransferLength - data_cnt;
		if(len > EP2_NONISO_PACKET_SIZE) len = EP2_NONISO_PACKET_SIZE;
		len = D12_WriteEndpoint(5, data_ptr + data_cnt, len);
		data_cnt += len;
		if(data_cnt == cbw.dCBWDataTransferLength) { /* M */
			/* ܑf[^̑MɁACSW𑗐M悤\񂵂܂B */
			phase = PHASE_CSW;
		}
		/*{{MO*/
		tx_cnt += len;
		/*}}MO*/
		return;

	case PHASE_CSW: /* CSWM҂ */
		D12_WriteEndpoint(5, &csw, CSW_SIZE);
		phase = PHASE_IDLE;
		return;
	}
}

void
fat_init()
{
	int retval, i;
	FILEACC fa;

	/* u[gZN^B */
	memset(&scts[0][0], 0, sizeof SCTSIZ);
	memcpy(&scts[0][0], &fat_boot_sector, sizeof fat_boot_sector);

	/* fBXNC[WA܂͏B */
	retval = pceFileOpen(&fa, USBSTOR_FILENAME, FOMD_RD);
	if(retval == 0 && fa.fsize == USBSTOR_FILESIZE) { /* Ô߃t@CTCY`FbN */
		for(i = 1/*u[gZN^͔΂*/; i < SCTCNT; i++) {
			pceFileReadSct(&fa, &scts[i][0], i - 1/*u[gZN^͕ۑĂȂ*/, SCTSIZ/*=BLOCKSIZE*/);
		}
	} else {
		/* FATB */
		memset(&scts[1][0], 0, sizeof SCTSIZ);
		scts[1][0] = 0xf8;
		scts[1][1] = 0xff;
		scts[1][2] = 0xff;

		/* fBNgB */
		memset(&scts[2][0], 0, sizeof SCTSIZ);
	}
}

void
fat_exit()
{
	int retval, i;
	FILEACC fa;

	/* ύXĂȂΕۑ܂B
	 * fat_init()ŃfBXNC[Wǂݍ܂ɏꍇAύXȂΕۑ܂B
	 * ȂȂȀ菇ŌiSȏԁj𕜌ł邩łB
	 */
	if(!dirty) return;

	/* fBXNC[WۑB */
	retval = pceFileCreate(USBSTOR_FILENAME, USBSTOR_FILESIZE); /* ɂu܂ */
	if(retval != 0) return; /* 󂫗eʕsH */
	retval = pceFileOpen(&fa, USBSTOR_FILENAME, FOMD_WR);
	if(retval == 0) { /* CreatêȂOpen͂AÔ */
		for(i = 1/*u[gZN^͔΂*/; i < SCTCNT; i++) {
			pceFileWriteSct(&fa, &scts[i][0], i - 1/*u[gZN^͕ۑȂ*/, SCTSIZ/*=BLOCKSIZE*/);
		}
	}
}

#if 0 /*Ă݂܂g܂łB얢mFB */
int
fat_get(int no)
{
	int ofs, val;

	if(no < CLSMIN || CLSMAX < no) return -1; /* ͈͊O */
	ofs = no / 2 * 3; /* 2GgyA3oCgÂ */

	val = scts[1][ofs + 2] << 16 |
	      scts[1][ofs + 1] <<  8 |
	      scts[1][ofs + 0] <<  0;
	if(no & 1) val >>= 12;
	val &= 0xfff;

	return val;
}
void
fat_set(int no, int val)
{
	int ofs, lo, hi;

	if(no < CLSMIN || CLSMAX < no) return; /* ͈͊O */
	ofs = no / 2 * 3; /* 2GgyA3oCgÂ */

	if(!(no & 1)) {
		lo = val & 0xfff;
		hi = fat_get(no |  1);
	} else {
		lo = fat_get(no & ~1);
		hi = val & 0xfff;
	}
	val = hi << 12 | lo;

	scts[1][ofs + 2] = (unsigned char)(val >> 16);
	scts[1][ofs + 1] = (unsigned char)(val >>  8);
	scts[1][ofs + 0] = (unsigned char)(val >>  0);
}
#endif
